use sprinklers_core::{
    model::{Program, ProgramId, ProgramItem, ProgramSequence, SectionId},
    schedule::Schedule,
};

use rusqlite::{
    types::{FromSql, FromSqlError, FromSqlResult, ToSqlOutput, Value, ValueRef},
    Error as SqlError, Result as SqlResult, Row as SqlRow, ToSql,
};
use serde::{Deserialize, Serialize};

pub struct SqlJson<T>(T);

impl<T> SqlJson<T> {
    pub fn into_inner(self) -> T {
        self.0
    }
}

impl<T> FromSql for SqlJson<T>
where
    for<'de> T: Deserialize<'de>,
{
    fn column_result(value: ValueRef<'_>) -> FromSqlResult<Self> {
        if let ValueRef::Text(text) = value {
            let deser_value: T =
                serde_json::from_slice(text).map_err(|err| FromSqlError::Other(Box::new(err)))?;
            Ok(SqlJson(deser_value))
        } else {
            Err(FromSqlError::InvalidType)
        }
    }
}

impl<T> ToSql for SqlJson<T>
where
    T: Serialize,
{
    fn to_sql(&self) -> SqlResult<ToSqlOutput<'_>> {
        serde_json::to_string(&self.0)
            .map(|serialized| ToSqlOutput::Owned(Value::Text(serialized)))
            .map_err(|err| SqlError::ToSqlConversionFailure(Box::new(err)))
    }
}

type SqlProgramSequence = SqlJson<ProgramSequence>;
type SqlSchedule = SqlJson<Schedule>;

pub fn from_sql<'a>(row: &SqlRow<'a>) -> SqlResult<Program> {
    Ok(Program {
        id: row.get(0)?,
        name: row.get(1)?,
        enabled: row.get(2)?,
        schedule: row.get::<_, SqlSchedule>(3)?.into_inner(),
        sequence: row.get::<_, SqlProgramSequence>(4)?.into_inner(),
    })
}

pub struct SqlProgram<'a> {
    id: ProgramId,
    name: &'a String,
    enabled: bool,
    schedule: SqlJson<&'a Schedule>,
}

impl<'a> IntoIterator for &'a SqlProgram<'a> {
    type Item = &'a dyn ToSql;
    type IntoIter = std::vec::IntoIter<Self::Item>;

    fn into_iter(self) -> Self::IntoIter {
        vec![
            &self.id as &dyn ToSql,
            self.name,
            &self.enabled,
            &self.schedule,
        ]
        .into_iter()
    }
}

pub fn as_sql(program: &Program) -> SqlProgram {
    SqlProgram {
        id: program.id,
        name: &program.name,
        enabled: program.enabled,
        schedule: SqlJson(&program.schedule),
    }
}

pub struct SqlProgramItem {
    program_id: ProgramId,
    seq_num: isize,
    section_id: SectionId,
    duration: f64,
}

impl<'a> IntoIterator for &'a SqlProgramItem {
    type Item = &'a dyn ToSql;
    type IntoIter = std::vec::IntoIter<Self::Item>;

    fn into_iter(self) -> Self::IntoIter {
        vec![
            &self.program_id as &dyn ToSql,
            &self.seq_num,
            &self.section_id,
            &self.duration,
        ]
        .into_iter()
    }
}

pub fn item_as_sql(
    program_item: &ProgramItem,
    program_id: ProgramId,
    seq_num: usize,
) -> SqlProgramItem {
    SqlProgramItem {
        program_id,
        seq_num: (seq_num + 1) as isize,
        section_id: program_item.section_id,
        duration: program_item.duration.as_secs_f64(),
    }
}

pub fn sequence_as_sql<'a>(program: &'a Program) -> impl Iterator<Item = SqlProgramItem> + 'a {
    let program_id = program.id;
    program
        .sequence
        .iter()
        .enumerate()
        .map(move |(seq_num, item)| item_as_sql(item, program_id, seq_num))
}
